/*********************************************************************/
/* Copyright (c) 2013, EPFL/Blue Brain Project                       */
/*                     Raphael Dumusc <raphael.dumusc@epfl.ch>       */
/* All rights reserved.                                              */
/*                                                                   */
/* Redistribution and use in source and binary forms, with or        */
/* without modification, are permitted provided that the following   */
/* conditions are met:                                               */
/*                                                                   */
/*   1. Redistributions of source code must retain the above         */
/*      copyright notice, this list of conditions and the following  */
/*      disclaimer.                                                  */
/*                                                                   */
/*   2. Redistributions in binary form must reproduce the above      */
/*      copyright notice, this list of conditions and the following  */
/*      disclaimer in the documentation and/or other materials       */
/*      provided with the distribution.                              */
/*                                                                   */
/*    THIS  SOFTWARE IS PROVIDED  BY THE  UNIVERSITY OF  TEXAS AT    */
/*    AUSTIN  ``AS IS''  AND ANY  EXPRESS OR  IMPLIED WARRANTIES,    */
/*    INCLUDING, BUT  NOT LIMITED  TO, THE IMPLIED  WARRANTIES OF    */
/*    MERCHANTABILITY  AND FITNESS FOR  A PARTICULAR  PURPOSE ARE    */
/*    DISCLAIMED.  IN  NO EVENT SHALL THE UNIVERSITY  OF TEXAS AT    */
/*    AUSTIN OR CONTRIBUTORS BE  LIABLE FOR ANY DIRECT, INDIRECT,    */
/*    INCIDENTAL,  SPECIAL, EXEMPLARY,  OR  CONSEQUENTIAL DAMAGES    */
/*    (INCLUDING, BUT  NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE    */
/*    GOODS  OR  SERVICES; LOSS  OF  USE,  DATA,  OR PROFITS;  OR    */
/*    BUSINESS INTERRUPTION) HOWEVER CAUSED  AND ON ANY THEORY OF    */
/*    LIABILITY, WHETHER  IN CONTRACT, STRICT  LIABILITY, OR TORT    */
/*    (INCLUDING NEGLIGENCE OR OTHERWISE)  ARISING IN ANY WAY OUT    */
/*    OF  THE  USE OF  THIS  SOFTWARE,  EVEN  IF ADVISED  OF  THE    */
/*    POSSIBILITY OF SUCH DAMAGE.                                    */
/*                                                                   */
/* The views and conclusions contained in the software and           */
/* documentation are those of the authors and should not be          */
/* interpreted as representing official policies, either expressed   */
/* or implied, of The University of Texas at Austin.                 */
/*********************************************************************/

#include "StateSerializationHelper.h"

#include "State.h"
#include "StatePreview.h"
#include "DisplayGroup.h"
#include "ContentFactory.h"

#include "globals.h"
#include "configuration/Configuration.h"

#include "log.h"

#include <fstream>
#include <boost/foreach.hpp>
#include <boost/archive/xml_iarchive.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_archive_exception.hpp>

StateSerializationHelper::StateSerializationHelper(DisplayGroupPtr displayGroup)
    : displayGroup_(displayGroup)
{
}

bool StateSerializationHelper::save(const QString& filename, const bool generatePreview)
{
    ContentWindowPtrs contentWindows = displayGroup_->getContentWindows();

    if (generatePreview)
    {
        const QSize wallDimensions(g_configuration->getTotalWidth(), g_configuration->getTotalHeight());
        StatePreview filePreview(filename);
        filePreview.generateImage( wallDimensions, contentWindows );
        filePreview.saveToFile();
    }

    // serialize state
    std::ofstream ofs(filename.toStdString());
    if (!ofs.good())
        return false;

    // brace this so destructor is called on archive before we use the stream
    {
        State state(contentWindows);
        boost::archive::xml_oarchive oa(ofs);
        oa << BOOST_SERIALIZATION_NVP(state);
    }
    ofs.close();

    return true;
}

bool StateSerializationHelper::load(const QString& filename)
{
    State state;

    // For backward compatibility, try to load the file as a legacy state file first
    if( state.legacyLoadXML( filename ))
    {
        displayGroup_->setContentWindows(state.getContentWindows());
        return true;
    }

    // De-serialize state file
    std::ifstream ifs(filename.toStdString());
    if (!ifs.good())
        return false;
    try
    {
        boost::archive::xml_iarchive ia(ifs);
        ia >> BOOST_SERIALIZATION_NVP(state);
    }
    catch(const boost::archive::archive_exception& e)
    {
        put_flog(LOG_ERROR, "Could not restore session/state %s: %s",
                 filename.toStdString().c_str(), e.what());
        return false;
    }
    ifs.close();

    ContentWindowPtrs contentWindows = state.getContentWindows();
    validate(contentWindows);

    displayGroup_->setContentWindows(contentWindows);
    return true;
}

void StateSerializationHelper::validate(ContentWindowPtrs& contentWindows) const
{
    ContentWindowPtrs validContentWindows;
    validContentWindows.reserve(contentWindows.size());

    BOOST_FOREACH(ContentWindowPtr contentWindow, contentWindows)
    {
        if (!contentWindow->getContent())
        {
            put_flog(LOG_WARN, "Window '%s' does not have a Content.",
                     contentWindow->getID().toString().toStdString().c_str());
            continue;
        }
        // PixelStreams are not supported yet, don't restore them.
        // This feature will be implemented in DISCL-6
        if (isPixelStream((contentWindow)))
            continue;

        finalize(contentWindow);
        validContentWindows.push_back(contentWindow);
    }
    contentWindows = validContentWindows;
}

bool StateSerializationHelper::isPixelStream(ContentWindowPtr contentWindow) const
{
    return contentWindow->getContent()->getType() == CONTENT_TYPE_PIXEL_STREAM;
}

void StateSerializationHelper::finalize(ContentWindowPtr contentWindow) const
{
    // Refresh content informations. Files can have changed since the State was saved.
    if (contentWindow->getContent()->readMetadata())
        contentWindow->createInteractionDelegate();
    else
        contentWindow->setContent(ContentFactory::getErrorContent());
}
